@@ -1,6 +1,6 @@
 /* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil -*- */
 /*
- * Copyright (c) 2014-2015 Intel, Inc.  All rights reserved.
+ * Copyright (c) 2014-2017 Intel, Inc. All rights reserved.
  * Copyright (c) 2014      Research Organization for Information Science
  *                         and Technology (RIST). All rights reserved.
  * Copyright (c) 2016      Mellanox Technologies, Inc.
@@ -45,11 +45,11 @@
 #define PMI_MAX_VAL_LEN      4096            /* Maximum size of a PMI value */
 
 
-#define PMI_CHECK() \
-	do {                     \
-        if (!pmi_init) {     \
-            return PMI_FAIL; \
-        }                    \
+#define PMI_CHECK()             \
+    do {                        \
+        if (!pmi_init) {        \
+            return PMI_FAIL;    \
+        }                       \
     } while (0)
 
 /* local functions */
@@ -57,6 +57,7 @@ static pmix_status_t convert_int(int *value, pmix_value_t *kv);
 static int convert_err(pmix_status_t rc);
 static pmix_proc_t myproc;
 static int pmi_init = 0;
+static bool pmi_singleton = false;
 
 PMIX_EXPORT int PMI_Init(int *spawned)
 {
@@ -66,7 +67,19 @@ PMIX_EXPORT int PMI_Init(int *spawned)
     pmix_info_t info[1];
     bool  val_optinal = 1;
 
-    if (PMIX_SUCCESS != PMIx_Init(&myproc)) {
+    if (PMIX_SUCCESS != (rc = PMIx_Init(&myproc))) {
+        /* if we didn't see a PMIx server (e.g., missing envar),
+         * then allow us to run as a singleton */
+        if (PMIX_ERR_INVALID_NAMESPACE == rc) {
+            if (NULL != spawned) {
+                *spawned = 0;
+            }
+            pmi_singleton = true;
+            (void)strncpy(myproc.nspace, "1234", PMIX_MAX_NSLEN);
+            myproc.rank = 0;
+            pmi_init = 1;
+            return PMI_SUCCESS;
+        }
         return PMI_ERR_INIT;
     }
 
@@ -109,7 +122,11 @@ PMIX_EXPORT int PMI_Initialized(PMI_BOOL *initialized)
         return PMI_ERR_INVALID_ARG;
     }
 
-    *initialized = (PMIx_Initialized() ? PMI_TRUE : PMI_FALSE);
+    if (pmi_singleton) {
+        *initialized = PMI_TRUE;
+    } else {
+        *initialized = (PMIx_Initialized() ? PMI_TRUE : PMI_FALSE);
+    }
 
     return PMI_SUCCESS;
 }
@@ -121,6 +138,10 @@ PMIX_EXPORT int PMI_Finalize(void)
     PMI_CHECK();
 
     pmi_init = 0;
+    if (pmi_singleton) {
+        return PMI_SUCCESS;
+    }
+
     rc = PMIx_Finalize();
     return convert_err(rc);
 }
@@ -131,6 +152,10 @@ PMIX_EXPORT int PMI_Abort(int flag, const char msg[])
 
     PMI_CHECK();
 
+    if (pmi_singleton) {
+        return PMI_SUCCESS;
+    }
+
     rc = PMIx_Abort(flag, msg, NULL, 0);
     return convert_err(rc);
 }
@@ -153,6 +178,9 @@ PMIX_EXPORT int PMI_KVS_Put(const char kvsname[], const char key[], const char v
     if ((value == NULL) || (strlen(value) > PMI_MAX_VAL_LEN)) {
         return PMI_ERR_INVALID_VAL;
     }
+    if (pmi_singleton) {
+        return PMI_SUCCESS;
+    }
 
     pmix_output_verbose(2, pmix_globals.debug_output,
             "PMI_KVS_Put: KVS=%s, key=%s value=%s", kvsname, key, value);
@@ -173,6 +201,9 @@ PMIX_EXPORT int PMI_KVS_Commit(const char kvsname[])
     if ((kvsname == NULL) || (strlen(kvsname) > PMI_MAX_KVSNAME_LEN)) {
         return PMI_ERR_INVALID_KVS;
     }
+    if (pmi_singleton) {
+        return PMI_SUCCESS;
+    }
 
     pmix_output_verbose(2, pmix_globals.debug_output, "PMI_KVS_Commit: KVS=%s",
             kvsname);
@@ -202,6 +233,30 @@ PMIX_EXPORT int PMI_KVS_Get( const char kvsname[], const char key[], char value[
     pmix_output_verbose(2, pmix_globals.debug_output,
             "PMI_KVS_Get: KVS=%s, key=%s value=%s", kvsname, key, value);
 
+    /* PMI-1 expects resource manager to set
+     * process mapping in ANL notation. */
+    if (!strcmp(key, ANL_MAPPING)) {
+        /* we are looking in the job-data. If there is nothing there
+         * we don't want to look in rank's data, thus set rank to widcard */
+        proc = myproc;
+        proc.rank = PMIX_RANK_WILDCARD;
+        if (PMIX_SUCCESS == PMIx_Get(&proc, PMIX_ANL_MAP, NULL, 0, &val) &&
+               (NULL != val) && (PMIX_STRING == val->type)) {
+            strncpy(value, val->data.string, length);
+            PMIX_VALUE_FREE(val, 1);
+            return PMI_SUCCESS;
+        } else {
+            /* artpol:
+             * Some RM's (i.e. SLURM) already have ANL precomputed. The export it
+             * through PMIX_ANL_MAP variable.
+             * If we haven't found it we want to have our own packing functionality
+             * since it's common.
+             * Somebody else has to write it since I've already done that for
+             * GPL'ed SLURM :) */
+            return PMI_FAIL;
+        }
+    }
+
     /* retrieve the data from PMIx - since we don't have a rank,
      * we indicate that by passing the UNDEF value */
     (void)strncpy(proc.nspace, kvsname, PMIX_MAX_NSLEN);
@@ -232,6 +287,10 @@ PMIX_EXPORT int PMI_Barrier(void)
 
     PMI_CHECK();
 
+    if (pmi_singleton) {
+        return PMI_SUCCESS;
+    }
+
     info = &buf;
     PMIX_INFO_CONSTRUCT(info);
     PMIX_INFO_LOAD(info, PMIX_COLLECT_DATA, &val, PMIX_BOOL);
@@ -247,9 +306,10 @@ PMIX_EXPORT int PMI_Get_size(int *size)
 {
     pmix_status_t rc = PMIX_SUCCESS;
     pmix_value_t *val;
-    pmix_proc_t proc;
     pmix_info_t info[1];
     bool  val_optinal = 1;
+    pmix_proc_t proc = myproc;
+    proc.rank = PMIX_RANK_WILDCARD;
 
     PMI_CHECK();
 
@@ -257,8 +317,10 @@ PMIX_EXPORT int PMI_Get_size(int *size)
         return PMI_ERR_INVALID_ARG;
     }
 
-    (void)strncpy(proc.nspace, myproc.nspace, PMIX_MAX_NSLEN);
-    proc.rank = PMIX_RANK_UNDEF;
+    if (pmi_singleton) {
+        *size = 1;
+        return PMI_SUCCESS;
+    }
 
     /* set controlling parameters
      * PMIX_OPTIONAL - expect that these keys should be available on startup
@@ -293,9 +355,10 @@ PMIX_EXPORT int PMI_Get_universe_size(int *size)
 {
     pmix_status_t rc = PMIX_SUCCESS;
     pmix_value_t *val;
-    pmix_proc_t proc;
     pmix_info_t info[1];
     bool  val_optinal = 1;
+    pmix_proc_t proc = myproc;
+    proc.rank = PMIX_RANK_WILDCARD;
 
     PMI_CHECK();
 
@@ -303,8 +366,10 @@ PMIX_EXPORT int PMI_Get_universe_size(int *size)
         return PMI_ERR_INVALID_ARG;
     }
 
-    (void)strncpy(proc.nspace, myproc.nspace, PMIX_MAX_NSLEN);
-    proc.rank = PMIX_RANK_UNDEF;
+    if (pmi_singleton) {
+        *size = 1;
+        return PMI_SUCCESS;
+    }
 
     /* set controlling parameters
      * PMIX_OPTIONAL - expect that these keys should be available on startup
@@ -327,9 +392,10 @@ PMIX_EXPORT int PMI_Get_appnum(int *appnum)
 {
     pmix_status_t rc = PMIX_SUCCESS;
     pmix_value_t *val;
-    pmix_proc_t proc;
     pmix_info_t info[1];
     bool  val_optinal = 1;
+    pmix_proc_t proc = myproc;
+    proc.rank = PMIX_RANK_WILDCARD;
 
     PMI_CHECK();
 
@@ -337,8 +403,10 @@ PMIX_EXPORT int PMI_Get_appnum(int *appnum)
         return PMI_ERR_INVALID_ARG;
     }
 
-    (void)strncpy(proc.nspace, myproc.nspace, PMIX_MAX_NSLEN);
-    proc.rank = PMIX_RANK_UNDEF;
+    if (pmi_singleton) {
+        *appnum = 0;
+        return PMI_SUCCESS;
+    }
 
     /* set controlling parameters
      * PMIX_OPTIONAL - expect that these keys should be available on startup
@@ -350,6 +418,10 @@ PMIX_EXPORT int PMI_Get_appnum(int *appnum)
     if (PMIX_SUCCESS == rc) {
         rc = convert_int(appnum, val);
         PMIX_VALUE_RELEASE(val);
+    } else if( PMIX_ERR_NOT_FOUND == rc ){
+        /* this is optional value, set to 0 */
+        *appnum = 0;
+        rc = PMIX_SUCCESS;
     }
 
     PMIX_INFO_DESTRUCT(&info[0]);
@@ -368,6 +440,10 @@ PMIX_EXPORT int PMI_Publish_name(const char service_name[], const char port[])
         return PMI_ERR_INVALID_ARG;
     }
 
+    if (pmi_singleton) {
+        return PMI_FAIL;
+    }
+
     /* pass the service/port */
     (void) strncpy(info.key, service_name, PMIX_MAX_KEYLEN);
     info.value.type = PMIX_STRING;
@@ -391,6 +467,10 @@ PMIX_EXPORT int PMI_Unpublish_name(const char service_name[])
         return PMI_ERR_INVALID_ARG;
     }
 
+    if (pmi_singleton) {
+        return PMI_FAIL;
+    }
+
     /* pass the service */
     keys[0] = (char*) service_name;
     keys[1] = NULL;
@@ -410,6 +490,10 @@ PMIX_EXPORT int PMI_Lookup_name(const char service_name[], char port[])
         return PMI_ERR_INVALID_ARG;
     }
 
+    if (pmi_singleton) {
+        return PMI_FAIL;
+    }
+
     PMIX_PDATA_CONSTRUCT(&pdata);
 
     /* pass the service */
@@ -481,6 +565,8 @@ PMIX_EXPORT int PMI_Get_clique_size(int *size)
     pmix_value_t *val;
     pmix_info_t info[1];
     bool  val_optinal = 1;
+    pmix_proc_t proc = myproc;
+    proc.rank = PMIX_RANK_WILDCARD;
 
     PMI_CHECK();
 
@@ -488,13 +574,18 @@ PMIX_EXPORT int PMI_Get_clique_size(int *size)
         return PMI_ERR_INVALID_ARG;
     }
 
+    if (pmi_singleton) {
+        *size = 1;
+        return PMI_SUCCESS;
+    }
+
     /* set controlling parameters
      * PMIX_OPTIONAL - expect that these keys should be available on startup
      */
     PMIX_INFO_CONSTRUCT(&info[0]);
     PMIX_INFO_LOAD(&info[0], PMIX_OPTIONAL, &val_optinal, PMIX_BOOL);
 
-    rc = PMIx_Get(&myproc, PMIX_LOCAL_SIZE, info, 1, &val);
+    rc = PMIx_Get(&proc, PMIX_LOCAL_SIZE, info, 1, &val);
     if (PMIX_SUCCESS == rc) {
         rc = convert_int(size, val);
         PMIX_VALUE_RELEASE(val);
@@ -511,6 +602,8 @@ PMIX_EXPORT int PMI_Get_clique_ranks(int ranks[], int length)
     pmix_value_t *val;
     char **rks;
     int i;
+    pmix_proc_t proc = myproc;
+    proc.rank = PMIX_RANK_WILDCARD;
 
     PMI_CHECK();
 
@@ -518,7 +611,12 @@ PMIX_EXPORT int PMI_Get_clique_ranks(int ranks[], int length)
         return PMI_ERR_INVALID_ARGS;
     }
 
-    rc = PMIx_Get(&myproc, PMIX_LOCAL_PEERS, NULL, 0, &val);
+    if (pmi_singleton) {
+        ranks[0] = 0;
+        return PMI_SUCCESS;
+    }
+
+    rc = PMIx_Get(&proc, PMIX_LOCAL_PEERS, NULL, 0, &val);
     if (PMIX_SUCCESS == rc) {
         /* kv will contain a string of comma-separated
          * ranks on my node */
@@ -629,6 +727,10 @@ PMIX_EXPORT int PMI_Spawn_multiple(int count,
         return PMI_ERR_INVALID_ARG;
     }
 
+    if (pmi_singleton) {
+        return PMI_FAIL;
+    }
+
     /* setup the apps */
     PMIX_APP_CREATE(apps, count);
     for (i = 0; i < count; i++) {
