@@ -5,17 +5,18 @@
  * Copyright (c) 2004-2005 The University of Tennessee and The University
  *                         of Tennessee Research Foundation.  All rights
  *                         reserved.
- * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart, 
+ * Copyright (c) 2004-2005 High Performance Computing Center Stuttgart,
  *                         University of Stuttgart.  All rights reserved.
  * Copyright (c) 2004-2005 The Regents of the University of California.
  *                         All rights reserved.
  * Copyright (c) 2007      Cisco Systems, Inc.  All rights reserved.
  * Copyright (c) 2011      Los Alamos National Security, LLC.  All rights
- *                         reserved. 
+ *                         reserved.
+ * Copyright (c) 2014      Intel Corporation.  All rights reserved.
  * $COPYRIGHT$
- * 
+ *
  * Additional copyrights may follow
- * 
+ *
  * $HEADER$
  */
 
@@ -26,9 +27,9 @@
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif  /* HAVE_UNISTD_H */
-#ifdef HAVE_STRING_H
 #include <string.h>
-#endif  /* HAVE_STRING_H */
+
+#include "opal/dss/dss.h"
 
 #include "orte/mca/rml/rml.h"
 #include "orte/mca/rml/rml_types.h"
@@ -53,9 +54,9 @@ void orte_iof_orted_send_xonxoff(orte_iof_tag_t tag)
 {
     opal_buffer_t *buf;
     int rc;
-    
+
     buf = OBJ_NEW(opal_buffer_t);
-    
+
     /* pack the tag - we do this first so that flow control messages can
      * consist solely of the tag
      */
@@ -72,7 +73,7 @@ void orte_iof_orted_send_xonxoff(orte_iof_tag_t tag)
 
     /* send the buffer to the HNP */
     if (0 > (rc = orte_rml.send_buffer_nb(ORTE_PROC_MY_HNP, buf, ORTE_RML_TAG_IOF_HNP,
-                                          0, send_cb, NULL))) {
+                                          send_cb, NULL))) {
         ORTE_ERROR_LOG(rc);
     }
 }
@@ -95,20 +96,20 @@ void orte_iof_orted_recv(int status, orte_process_name_t* sender,
     orte_process_name_t target;
     opal_list_item_t *item;
     int rc;
-    
+
     /* see what stream generated this data */
     count = 1;
     if (ORTE_SUCCESS != (rc = opal_dss.unpack(buffer, &stream, &count, ORTE_IOF_TAG))) {
         ORTE_ERROR_LOG(rc);
         goto CLEAN_RETURN;
     }
-    
+
     /* if this isn't stdin, then we have an error */
     if (ORTE_IOF_STDIN != stream) {
         ORTE_ERROR_LOG(ORTE_ERR_COMM_FAILURE);
         goto CLEAN_RETURN;
     }
-    
+
     /* unpack the intended target */
     count = 1;
     if (ORTE_SUCCESS != (rc = opal_dss.unpack(buffer, &target, &count, ORTE_NAME))) {
@@ -123,18 +124,22 @@ void orte_iof_orted_recv(int status, orte_process_name_t* sender,
         goto CLEAN_RETURN;
     }
     /* numbytes will contain the actual #bytes that were sent */
-    
+
     OPAL_OUTPUT_VERBOSE((1, orte_iof_base_framework.framework_output,
                          "%s unpacked %d bytes for local proc %s",
                          ORTE_NAME_PRINT(ORTE_PROC_MY_NAME), numbytes,
                          ORTE_NAME_PRINT(&target)));
-    
+
     /* cycle through our list of sinks */
     for (item = opal_list_get_first(&mca_iof_orted_component.sinks);
          item != opal_list_get_end(&mca_iof_orted_component.sinks);
          item = opal_list_get_next(item)) {
         orte_iof_sink_t* sink = (orte_iof_sink_t*)item;
-        
+
+        /* is this sink for stdin? it could be for writing a file from stdout/err */
+        if (!(ORTE_IOF_STDIN & sink->tag))
+            continue;
+
         /* is this intended for this jobid? */
         if (target.jobid == sink->name.jobid) {
             /* yes - is this intended for all vpids or this vpid? */
